classdef monogenic
    % MONOGENIC monogenic class constructor
    % obj=monogenic(signal,method,params)
    % computes the monogenic scale-space
    %
    % signal: can be a scalar array (int, single, double) or a complex array
    % (interpreted as the Riesz transform of some signal)
    % method: method to compute the monogenic signal; possible choices (def: 'get'):
    %
    % 'dct' for a finite domain eigenfunction solution; parameters
    % params.scales: a vector of scales in Poisson scale space
    % params.coeffs: a vector of linear coefficients to compose e.g. a bandpass
    % filter; default: allpass
    %
    % 'sqf', 'sqfs', 'sqfx', 'sqfy' for a convolution solution (spherical quadrature
    % filter); parameters
    % params.scales: a 2-vector of scales in Poisson scale space (default [1 2])
    % params.trunce: truncation error in percentage (default 5%)
    %
    % 'get', 'gets' for implementation based on gradient energy tensor; parameters
    % params.scales: a scalar with standard deviation for Gaussian
    % derivative; default: Sobel filter
    %
    % member functions:
    %   disp
    %   size
    %   subsref
    %   subsasgn
    %   end
    %   log
    %   amplit
    %   orient
    %   phase
    %   recon
    %   conj
    

    % $Id: monogenic.m 26114 2012-06-26 08:48:06Z mfe $

    % Copyright (c) 2006 Michael Felsberg, mfe@isy.liu.se
    %
    %    This program is free software; you can redistribute it and/or modify
    %    it under the terms of the GNU General Public License as published by
    %    the Free Software Foundation; either version 2 of the License, or
    %    (at your option) any later version.
    %
    %    This program is distributed in the hope that it will be useful,
    %    but WITHOUT ANY WARRANTY; without even the implied warranty of
    %    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    %    GNU General Public License for more details.
    %
    %    You should have received a copy of the GNU General Public License
    %    along with this program; if not, write to the Free Software
    %    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    properties
        P=[];
        Q=[];
    end

    methods
        function obj=monogenic(signal,method,params)
            % MONOGENIC constructor
            if nargin==0	% standard
                obj.P=[];
                obj.Q=[];
            elseif nargin==1 && isa(signal,'monogenic')	% copy
                obj=signal;
            else    % generic
                if nargin==1
                    method='get';	%default
                end
                switch method(1:3)
                    case 'dct'
                        F=dct2(signal);
                        [sy sx]=size(F);
                        [x y]=meshgrid(0:sx-1,0:sy-1);
                        x=x./sx;
                        y=y./sy;
                        z=x+1i*y;
                        if numel(method)==4
                            switch method(4)
                                case 's'
                                    F=F.*abs(z);
                                case 'x' % buggy -avoid!!
                                    F=F.*x;
                                case 'y' % buggy -avoid!!
                                    F=F.*y;
                                otherwise
                                    error('Unknown method');
                            end
                        end
                        Rx=(x+1/sx)./abs(1/sx+z);
                        Ry=(y+1/sy)./abs(1i/sy+z);
                        w=-pi*sqrt(x.^2+y.^2); %abs(z)
                        Fx=Rx.*[F(:,2:end) zeros(size(F,1),1)];
                        Fy=Ry.*[F(2:end,:); zeros(1,size(F,2))];
                        if nargin==3
                            if ~isfield(params,'scales')
                                error('Wrong parameter struct.');
                            end
                            if isfield(params,'coeffs')
                                if length(params.scales)~=length(params.coeffs)
                                    error('Different dimensions of parameters.');
                                end
                                H=zeros(size(w));
                                for k=1:length(params.scales(:))
                                    H=H+params.coeffs(k)*exp(w*params.scales(k));
                                end
                            else
                                if length(params.scales)~=2
                                    error(['Wrong number of scale parameters, '...
                                        'should be two (center frequency and' ...
                                        'relative bandwidth)']);
                                end
                                H=exp(-4/log(2)/params.scales(2)^2*(log(-w/params.scales(1))).^2); % lognormal filter
                            end
                            Hx=[H(:,2:end) zeros(size(H,1),1)];
                            Hy=[H(2:end,:); zeros(1,size(H,2))];
                            obj.P=idct2(F.*H);
                            obj.Q=idst(idct(Fx.*Hx)')'+1i*idct(idst(Fy.*Hy)')';
                            obj.Q=obj.Q-mean(obj.Q(:)); %?
                        elseif nargin<3
                            obj.P=idct2(F);
                            obj.Q=idst(idct(Fx)')'+1i*idct(idst(Fy)')';
                            obj.Q=obj.Q-mean(obj.Q(:)); %?
                        else
                            error('Unknown number of parameters');
                        end
                    case 'sqf'
                        if nargin<3
                            s1=1;
                            s2=2;
                            e=5;
                        elseif nargin==3
                            if isfield(params,'scales') && numel(params.scales)==2
                                s1=min(params.scales);
                                s2=max(params.scales);
                            else
                                error('Wrong parameter struct.');
                            end
                            if isfield(params,'trunce')
                                e=params.trunce;
                            else
                                e=5;
                            end
                        else
                            error('Unknown number of parameters');
                        end
                        eval(['[b,c]=' strrep(method,'sqf','create_DOP') '(s1,s2,e);']);
                        obj.P=conv2(double(signal),b,'same');
                        obj.Q=conv2(double(signal),c,'same');
                    case 'get'
                        if nargin<3 % Sobel GET
                            [obj.P,obj.Q]=sGET(double(signal));
                        elseif nargin==3
                            if isfield(params,'scales') && numel(params.scales)==1
                                %[obj.P,obj.Q]=gGET(double(signal),params.scales);
                                eval(['[obj.P,obj.Q]=' strrep(method,'get','gGET') '(double(signal),params.scales);']);
                            else
                                error('Wrong parameter struct.');
                            end
                        else
                            error('Unknown number of parameters');
                        end
                    otherwise
                        error('Unknown method');
                end
            end
        end

        function disp(obj)
            % DISP(M)
            %
            % Displays
            if length(obj.P)>1
                disp(['monogenic signal of size ' num2str(size(obj.P,1)) 'x' ...
                    num2str(size(obj.P,2))]);
                disp(' ');
            else
                disp([obj.P real(obj.Q) imag(obj.Q)]);
            end
        end

        function D=size(obj)
            D=size(obj.P);
        end

        function m=subsref(obj,S)
            switch S(1).type
                case '()'
                    m=monogenic;
                    m.P=obj.P(S(1).subs{:});
                    m.Q=obj.Q(S(1).subs{:});
                case '.'
                    switch S(1).subs
                        case {'even', 'image', 'atten', 'signal'}
                            m=obj.P;
                        case {'odd', 'flux', 'phase', 'riesz'}
                            m=obj.Q;
                        otherwise
                            error('no such field');
                    end
                otherwise
                    error('no such subsref');
            end
            if numel(S)>1
                m=subsref(m,S(2:end));
            end
        end

        function obj=subsasgn(obj,S,m)
            switch S(end).type
                case '()'
                    if numel(S)==1
                        m=monogenic(m);
                        obj.P(S.subs{:})=m.P;
                        obj.Q(S.subs{:})=m.Q;
                    else
                        objt=subsref(obj,S(2));
                        objt=subsasgn(objt,S(1),m);
                        obj=subsasgn(obj,S(2),objt);
                    end
                case '.'
                    if numel(S)==1
                        if isnumeric(m) && size(m,1)==size(obj.P,1) && size(m,2)==size(obj.P,2)
                            switch S.subs
                                case {'even', 'image', 'atten', 'signal'}
                                    obj.P=m;
                                case {'odd', 'flux', 'phase', 'riesz'}
                                    obj.Q=m;
                                otherwise
                                    error('no such field');
                            end
                        elseif ~isnumeric(m)
                            error('wrong rvalue type');
                        else
                            error('wrong rvalue dimension');
                        end
                    else
                        obj=subsasgn(obj,S(1),subsasgn(subsref(obj,S(1)),S(2),m));
                    end
                otherwise
                    error('no such subsasgn');
            end
        end

        function nobj=log(obj)
            nobj=obj;
            nobj.P=log(abs(obj.P).^2+abs(obj.Q).^2)/2; % +1 removed !!
            warning off MATLAB:divideByZero
            if isreal(obj.P)
                sP=sign(obj.P);
            else
                error('to be implemented');
            end
            nobj.Q=obj.Q./abs(obj.Q).*mod(atan(sP.*abs(obj.Q)./abs(obj.P)),pi);
        end

        function a=amplit(obj)
            dum=log(obj);
            a=exp(dum.P);
        end

        function o=orient(obj)
            o=angle(obj.Q.^2)/2;
        end

        function p=phase(obj)
            p=angle(obj.P+1i*sign(real(obj.Q)).*abs(obj.Q));
        end

        function r=recon(obj)
            r=(exp(real(obj.P))).*cos(abs(obj.Q));
            r=r-mean(r(:));
        end

        function obj=conj(obj)
            dum=obj.P;
            obj.P=-conj(obj.Q);
            obj.Q=conj(dum);
        end

        function l=end(obj,k,n)
            if n==1
                l=numel(obj.P);
            elseif n==2
                l=size(obj.P,k);
            else
                error('unsupported number of dimensions');
            end
        end

    end

end